@@ -64,7 +64,7 @@ class TestConfiguration:
                     'prefixes': ['10/8'],
                     'asns': ['23456', '10-20']
                 },
-                'access-lists': {
+                'access_lists': {
                     'valid-list': {
                         '192/24',
                         '192.0.2.1',
@@ -87,6 +87,10 @@ class TestConfiguration:
                         'import_serial_source': 'ftp://example.com/serial',
                         'keep_journal': True,
                     },
+                    'TESTDB3': {
+                        'export_destination_unfiltered': '/tmp',
+                        'nrtm_access_list_unfiltered': 'valid-list',
+                    },
                     # RPKI source permitted, rpki.roa_source not set
                     'RPKI': {},
                 },
@@ -175,7 +179,7 @@ class TestConfiguration:
                     'from': 'example@example.com',
                     'smtp': '192.0.2.1'
                 },
-                'access-lists': {
+                'access_lists': {
                     'valid-list': {
                         '192/24',
                         '192.0.2.1',
@@ -217,13 +221,15 @@ class TestConfiguration:
     def test_load_invalid_config(self, save_yaml_config, tmpdir):
         config = {
             'irrd': {
+                'database_readonly': True,
                 'piddir': str(tmpdir + '/does-not-exist'),
+                'user': 'a',
                 'server': {
                     'whois': {
                         'access_list': 'doesnotexist',
                     },
                     'http': {
-                        'access_list': ['foo'],
+                        'status_access_list': ['foo'],
                     },
                 },
                 'email': {
@@ -252,15 +258,18 @@ class TestConfiguration:
                         'import_timer': 'foo',
                         'export_timer': 'bar',
                         'nrtm_host': '192.0.2.1',
+                        'unknown': True,
                     },
                     'TESTDB2': {
                         'authoritative': True,
                         'nrtm_host': '192.0.2.1',
                         'nrtm_port': 'not a number',
+                        'nrtm_access_list': 'invalid-list',
                     },
                     'TESTDB3': {
                         'authoritative': True,
                         'import_source': '192.0.2.1',
+                        'nrtm_access_list_unfiltered': 'invalid-list',
                     },
                     # Not permitted, rpki.roa_source is set
                     'RPKI': {},
@@ -270,7 +279,9 @@ class TestConfiguration:
                 'log': {
                     'level': 'INVALID',
                     'logging_config_path': 'path',
-                }
+                    'unknown': True,
+                },
+                'unknown_setting': False,
             }
         }
 
@@ -284,9 +295,10 @@ class TestConfiguration:
         assert 'Setting email.smtp is required.' in str(ce.value)
         assert 'Setting email.footer must be a string, if defined.' in str(ce.value)
         assert 'Setting email.recipient_override must be an email address if set.' in str(ce.value)
+        assert 'Settings user and group must both be defined, or neither.' in str(ce.value)
         assert 'Setting auth.gnupg_keyring is required.' in str(ce.value)
-        assert 'Access lists doesnotexist referenced in settings, but not defined.' in str(ce.value)
-        assert 'Setting server.http.access_list must be a string, if defined.' in str(ce.value)
+        assert 'Access lists doesnotexist, invalid-list referenced in settings, but not defined.' in str(ce.value)
+        assert 'Setting server.http.status_access_list must be a string, if defined.' in str(ce.value)
         assert 'Invalid item in access list bad-list: IPv4 Address with more than 4 bytes.' in str(ce.value)
         assert 'Invalid item in prefix scopefilter: invalid-prefix' in str(ce.value)
         assert 'Invalid item in asn scopefilter: invalid.' in str(ce.value)
@@ -296,6 +308,8 @@ class TestConfiguration:
         assert 'Setting nrtm_host for source TESTDB can not be enabled without setting import_serial_source.' in str(ce.value)
         assert 'Setting authoritative for source TESTDB2 can not be enabled when either nrtm_host or import_source are set.' in str(ce.value)
         assert 'Setting authoritative for source TESTDB3 can not be enabled when either nrtm_host or import_source are set.' in str(ce.value)
+        assert 'Source TESTDB can not have authoritative, import_source or nrtm_host set when database_readonly is enabled.' in str(ce.value)
+        assert 'Source TESTDB3 can not have authoritative, import_source or nrtm_host set when database_readonly is enabled.' in str(ce.value)
         assert 'Setting nrtm_port for source TESTDB2 must be a number.' in str(ce.value)
         assert 'Setting rpki.roa_import_timer must be set to a number.' in str(ce.value)
         assert 'Setting rpki.notify_invalid_subject must be a string, if defined.' in str(ce.value)
@@ -308,6 +322,9 @@ class TestConfiguration:
         assert 'Setting sources_default contains unknown sources: DOESNOTEXIST-DB' in str(ce.value)
         assert 'Invalid log.level: INVALID' in str(ce.value)
         assert 'Setting log.logging_config_path can not be combined' in str(ce.value)
+        assert 'Unknown setting key: unknown_setting' in str(ce.value)
+        assert 'Unknown setting key: log.unknown' in str(ce.value)
+        assert 'Unknown key(s) under source TESTDB: unknown' in str(ce.value)
 
 
 class TestGetSetting:
@@ -321,3 +338,11 @@ class TestGetSetting:
     def test_get_setting_env(self, monkeypatch):
         monkeypatch.setenv(self.env_name, 'env_value')
         assert get_setting(self.setting_name) == 'env_value'
+
+    def test_get_setting_unknown(self, monkeypatch):
+        with pytest.raises(ValueError):
+            get_setting('unknown')
+        with pytest.raises(ValueError):
+            get_setting('log.unknown')
+        with pytest.raises(ValueError):
+            get_setting('sources.TEST.unknown')
