@@ -45,6 +45,7 @@ class BulkRouteROAValidator:
     11000000000000000000001000, and therefore will (and should)
     not be included in the validation process.
     """
+
     def __init__(self, dh: DatabaseHandler, roas: Optional[List[ROA]] = None):
         """
         Create a validator object. Can use either a list of ROA objects,
@@ -85,7 +86,7 @@ class BulkRouteROAValidator:
         validation result, are not included in the return value.
         """
         columns = ['pk', 'rpsl_pk', 'ip_first', 'prefix_length', 'asn_first', 'source',
-                   'object_class', 'object_text', 'rpki_status']
+                   'rpki_status', 'scopefilter_status']
         q = RPSLDatabaseQuery(column_names=columns, enable_ordering=False)
         q = q.object_classes(['route', 'route6'])
         if sources:
@@ -107,6 +108,15 @@ class BulkRouteROAValidator:
                 result['rpki_status'] = new_status
                 objs_changed[new_status].append(result)
 
+        # Object text and class are only retrieved for objects with state changes
+        pks_to_enrich = [obj['pk'] for objs in objs_changed.values() for obj in objs]
+        query = RPSLDatabaseQuery(['pk', 'object_text', 'object_class'], enable_ordering=False).pks(pks_to_enrich)
+        rows_per_pk = {row['pk']: row for row in self.database_handler.execute_query(query)}
+
+        for rpsl_objs in objs_changed.values():
+            for rpsl_obj in rpsl_objs:
+                rpsl_obj.update(rows_per_pk[rpsl_obj['pk']])
+
         return objs_changed[RPKIStatus.valid], objs_changed[RPKIStatus.invalid], objs_changed[RPKIStatus.not_found]
 
     def validate_route(self, prefix_ip: str, prefix_length: int, prefix_asn: int, source: str) -> RPKIStatus:
