@@ -1,14 +1,17 @@
 import datetime
 import re
 from typing import List, Type, Optional
+from urllib.parse import urlparse
 
 from IPy import IP
 
+from irrd.utils.text import clean_ip_value_error
+from irrd.utils.validators import parse_as_number, ValidationError
 from .config import PASSWORD_HASHERS
 from .parser_state import RPSLParserMessages, RPSLFieldParseResult
-from irrd.utils.validators import parse_as_number, ValidationError
 
 # The IPv4/IPv6 regexes are for initial screening - not full validators
+
 re_ipv4_prefix = re.compile(r"^\d+\.\d+\.\d+\.\d+/\d+$")
 re_ipv6_prefix = re.compile(r"^[A-F\d:]+/\d+$", re.IGNORECASE)
 
@@ -25,9 +28,6 @@ reserved_words = ["ANY", "AS-ANY", "RS_ANY", "PEERAS", "AND", "OR", "NOT", "ATOM
                   "ACCEPT", "ANNOUNCE", "EXCEPT", "REFINE", "NETWORKS", "INTO", "INBOUND", "OUTBOUND"]
 reserved_prefixes = ["AS-", "RS-", "RTRS-", "FLTR-", "PRNG-"]
 
-# Turn "IP('193.0.1.1/21') has invalid prefix length (21)" into "invalid prefix length (21)"
-re_clean_ip_error = re.compile(r"IP\('[A-F0-9:./]+'\) has ", re.IGNORECASE)
-
 """
 Fields for RPSL data.
 
@@ -52,8 +52,13 @@ class RPSLTextField:
     If keep_case is False, all data is converted to upper case. The original object text is not
     modified for this. As parsed_data is indexed in the databases, this is important for searches,
     to match e.g. 'mntner: FOO' to 'mnt-by: foo' - as these values are equivalent.
+
+    The extracts property defines which fields from RPSLFieldParseResult are extracted
+    by a field, other than the value. The parser only consider these extractions
+    for primary key or lookup key fields.
     """
     keep_case = True
+    extracts: List[str] = []
 
     def __init__(self, optional: bool=False, multiple: bool=False, primary_key: bool=False, lookup_key: bool=False) -> None:
         self.optional = optional
@@ -91,6 +96,8 @@ class RPSLFieldListMixin:
 
 class RPSLIPv4PrefixField(RPSLTextField):
     """Field for a single IPv4 prefix."""
+    extracts = ['ip_first', 'ip_last', 'prefix', 'prefix_length']
+
     def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
         if not re_ipv4_prefix.match(value):
             messages.error(f'Invalid address prefix: {value}')
@@ -119,6 +126,8 @@ class RPSLIPv4PrefixesField(RPSLFieldListMixin, RPSLIPv4PrefixField):
 
 class RPSLIPv6PrefixField(RPSLTextField):
     """Field for a single IPv6 prefix."""
+    extracts = ['ip_first', 'ip_last', 'prefix', 'prefix_length']
+
     def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
         if not re_ipv6_prefix.match(value):
             messages.error(f'Invalid address prefix: {value}')
@@ -152,6 +161,8 @@ class RPSLIPv4AddressRangeField(RPSLTextField):
     Note that a single IP address is also valid, and that the range does
     not have to align to bitwise boundaries of prefixes.
     """
+    extracts = ['ip_first', 'ip_last']
+
     def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
         value = value.replace(',', '')  # #311, process multiline PK correctly
         if '-' in value:
@@ -263,6 +274,8 @@ class RPSLRouteSetMembersField(RPSLFieldListMixin, RPSLRouteSetMemberField):
 
 class RPSLASNumberField(RPSLTextField):
     """Field for a single AS number (in ASxxxx syntax)."""
+    extracts = ['asn']
+
     def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
         try:
             parsed_str, parsed_int = parse_as_number(value)
@@ -276,6 +289,8 @@ class RPSLASNumberField(RPSLTextField):
 
 class RPSLASBlockField(RPSLTextField):
     """Field for a block of AS numbers, e.g. AS1 - AS5."""
+    extracts = ['asn_first', 'asn_last']
+
     def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
         if '-' not in value:
             messages.error(f'Invalid AS range: {value}: does not contain a hyphen')
@@ -311,7 +326,7 @@ class RPSLSetNameField(RPSLTextField):
     must be a valid set name for this specific set, i.e. start with the given prefix.
 
     The prefix provided is the expected prefix of the set name, e.g. 'RS' for
-    a route-set, or 'AS' for an as-set.R
+    a route-set, or 'AS' for an as-set.
     """
     keep_case = False
 
@@ -365,6 +380,19 @@ class RPSLDNSNameField(RPSLTextField):
         return RPSLFieldParseResult(value)
 
 
+class RPSLURLField(RPSLTextField):
+    """Field for a URL, as used in e.g. geofeed attribute."""
+    keep_case = False
+    permitted_schemes = ['http', 'https']
+
+    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
+        result = urlparse(value)
+        if all([result.scheme in self.permitted_schemes, result.netloc]):
+            return RPSLFieldParseResult(value)
+        messages.error(f'Invalid {"/".join(self.permitted_schemes)} URL: {value}')
+        return None
+
+
 class RPSLGenericNameField(RPSLTextField):
     """
     Field for a generic name.
@@ -439,7 +467,7 @@ class RPSLReferenceField(RPSLTextField):
 
     def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
         if not self.referring_identifier_fields:
-            self._build_cache()
+            self.resolve_references()
 
         referring_field_messages = RPSLParserMessages()
         for identifier_field in self.referring_identifier_fields:
@@ -451,7 +479,7 @@ class RPSLReferenceField(RPSLTextField):
         messages.merge_messages(referring_field_messages)
         return None
 
-    def _build_cache(self):
+    def resolve_references(self):
         from .rpsl_objects import OBJECT_CLASS_MAPPING
         for ref in self.referring:
             rpsl_object_class = OBJECT_CLASS_MAPPING[ref]
@@ -534,7 +562,3 @@ def parse_set_name(prefixes: List[str], value: str, messages: RPSLParserMessages
     if parsed_value != value:
         messages.info(f'Set name {value} was reformatted as {parsed_value}')
     return RPSLFieldParseResult(parsed_value)
-
-
-def clean_ip_value_error(value_error):
-    return re.sub(re_clean_ip_error, '', str(value_error))
