@@ -23,6 +23,7 @@ from .queries import (BaseRPSLObjectDatabaseQuery, DatabaseStatusQuery,
 
 logger = logging.getLogger(__name__)
 MAX_RECORDS_BUFFER_BEFORE_INSERT = 15000
+RPSLDatabaseResponse = Iterator[Dict[str, Any]]
 
 
 class DatabaseHandler:
@@ -49,8 +50,12 @@ class DatabaseHandler:
 
         If readonly is True, this instance will expect read queries only.
         No transaction will be started, all queries will use autocommit.
+        Readonly is always true if database_readonly is set in the config.
         """
-        self.readonly = readonly
+        if get_setting('database_readonly'):
+            self.readonly = True
+        else:
+            self.readonly = readonly
         self.journaling_enabled = not readonly
         self._connection = get_engine().connect()
         if self.readonly:
@@ -102,6 +107,7 @@ class DatabaseHandler:
         """
         Commit any pending changes to the database and start a fresh transaction.
         """
+        self._check_write_permitted()
         self._flush_rpsl_object_writing_buffer()
         self._flush_roa_writing_buffer()
         self.status_tracker.finalise_transaction()
@@ -127,18 +133,35 @@ class DatabaseHandler:
     def execute_query(self,
                       query: Union[BaseRPSLObjectDatabaseQuery, DatabaseStatusQuery, RPSLDatabaseObjectStatisticsQuery, ROADatabaseObjectQuery],
                       flush_rpsl_buffer=True,
-                      ) -> Iterator[Dict[str, Any]]:
+                      refresh_on_error=False,
+                      ) -> RPSLDatabaseResponse:
         """
         Execute an RPSLDatabaseQuery within the current transaction.
         If flush_rpsl_buffer is set, the RPSL object buffer is flushed first.
+        If refresh_on_error is set, if any exception occurs, will refresh
+        the connection and retry.
         """
-        # To be able to query objects that were just created, flush the buffer.
-        if not self.readonly and flush_rpsl_buffer:
-            self._flush_rpsl_object_writing_buffer()
-        statement = query.finalise_statement()
-        result = self._connection.execute(statement)
-        for row in result.fetchall():
-            yield dict(row)
+        def execute_query():
+            # To be able to query objects that were just created, flush the buffer.
+            if not self.readonly and flush_rpsl_buffer:
+                self._flush_rpsl_object_writing_buffer()
+            statement = query.finalise_statement()
+            return self._connection.execute(statement)
+
+        try:
+            result = execute_query()
+        except Exception as exc:  # pragma: no cover
+            if refresh_on_error:
+                self.refresh_connection()
+                result = execute_query()
+            else:
+                raise exc
+
+        result_partition = result.fetchmany()
+        while result_partition:
+            for row in result_partition:
+                yield dict(row)
+            result_partition = result.fetchmany()
         result.close()
 
     def execute_statement(self, statement):
@@ -168,6 +191,7 @@ class DatabaseHandler:
         PK is unique in the database. This essentially only applies to inserting
         RPKI psuedo-IRR objects.
         """
+        self._check_write_permitted()
         if not rpsl_guaranteed_no_existing:
             self._rpsl_guaranteed_no_existing = False
         ip_first = str(rpsl_object.ip_first) if rpsl_object.ip_first else None
@@ -183,6 +207,7 @@ class DatabaseHandler:
         # seen before, the buffer must be flushed right away, or the two updates
         # will conflict.
         source = rpsl_object.parsed_data['source']
+
         rpsl_pk_source = rpsl_object.pk() + '-' + source
         if rpsl_pk_source in self._rpsl_pk_source_seen:
             self._flush_rpsl_object_writing_buffer()
@@ -198,6 +223,7 @@ class DatabaseHandler:
             'ip_first': ip_first,
             'ip_last': ip_last,
             'ip_size': ip_size,
+            'prefix': str(rpsl_object.prefix) if rpsl_object.prefix else None,
             'prefix_length': rpsl_object.prefix_length,
             'asn_first': rpsl_object.asn_first,
             'asn_last': rpsl_object.asn_last,
@@ -222,6 +248,7 @@ class DatabaseHandler:
         as they use a single COPY command. This is possible because
         ROA objects are never updated, only inserted.
         """
+        self._check_write_permitted()
         self._roa_insert_buffer.append({
             'ip_version': ip_version,
             'prefix': prefix_str,
@@ -241,6 +268,7 @@ class DatabaseHandler:
         entry, so that mirrors follow the (in)visibility depending
         on RPKI status.
         """
+        self._check_write_permitted()
         table = RPSLDatabaseObject.__table__
         if rpsl_objs_now_valid:
             pks = {o['pk'] for o in rpsl_objs_now_valid}
@@ -256,7 +284,10 @@ class DatabaseHandler:
             self.execute_statement(stmt)
 
         for rpsl_obj in rpsl_objs_now_valid + rpsl_objs_now_not_found:
-            if rpsl_obj['old_status'] == RPKIStatus.invalid:
+            if all([
+                rpsl_obj['old_status'] == RPKIStatus.invalid,
+                rpsl_obj['scopefilter_status'] == ScopeFilterStatus.in_scope
+            ]):
                 self.status_tracker.record_operation(
                     operation=DatabaseOperation.add_or_update,
                     rpsl_pk=rpsl_obj['rpsl_pk'],
@@ -290,6 +321,7 @@ class DatabaseHandler:
         entry, so that mirrors follow the (in)visibility depending
         on scopefilter status.
         """
+        self._check_write_permitted()
         table = RPSLDatabaseObject.__table__
         if rpsl_objs_now_in_scope:
             pks = {o['rpsl_pk'] for o in rpsl_objs_now_in_scope}
@@ -305,16 +337,17 @@ class DatabaseHandler:
             self.execute_statement(stmt)
 
         for rpsl_obj in rpsl_objs_now_in_scope:
-            self.status_tracker.record_operation(
-                operation=DatabaseOperation.add_or_update,
-                rpsl_pk=rpsl_obj['rpsl_pk'],
-                source=rpsl_obj['source'],
-                object_class=rpsl_obj['object_class'],
-                object_text=rpsl_obj['object_text'],
-                origin=JournalEntryOrigin.scope_filter,
-                source_serial=None,
-            )
-            self._object_classes_modified.add(rpsl_obj['object_class'])
+            if rpsl_obj['rpki_status'] != RPKIStatus.invalid:
+                self.status_tracker.record_operation(
+                    operation=DatabaseOperation.add_or_update,
+                    rpsl_pk=rpsl_obj['rpsl_pk'],
+                    source=rpsl_obj['source'],
+                    object_class=rpsl_obj['object_class'],
+                    object_text=rpsl_obj['object_text'],
+                    origin=JournalEntryOrigin.scope_filter,
+                    source_serial=None,
+                )
+                self._object_classes_modified.add(rpsl_obj['object_class'])
 
         for rpsl_obj in rpsl_objs_now_out_scope_as + rpsl_objs_now_out_scope_prefix:
             if rpsl_obj['old_status'] == ScopeFilterStatus.in_scope:
@@ -338,6 +371,7 @@ class DatabaseHandler:
         The origin indicates the origin of this change, see JournalEntryOrigin
         for the various options.
         """
+        self._check_write_permitted()
         self._flush_rpsl_object_writing_buffer()
         table = RPSLDatabaseObject.__table__
         if not source and rpsl_object:
@@ -385,7 +419,9 @@ class DatabaseHandler:
         if not self._rpsl_upsert_buffer:
             return
 
-        rpsl_composite_key = ['rpsl_pk', 'source']
+        self._check_write_permitted()
+
+        rpsl_composite_key = ['rpsl_pk', 'source', 'object_class']
         stmt = pg.insert(RPSLDatabaseObject).values([x[0] for x in self._rpsl_upsert_buffer])
 
         if not self._rpsl_guaranteed_no_existing:
@@ -438,6 +474,8 @@ class DatabaseHandler:
         if not self._roa_insert_buffer:
             return
 
+        self._check_write_permitted()
+
         columns = list(self._roa_insert_buffer[0].keys())
         roa_rows = []
         for roa in self._roa_insert_buffer:
@@ -457,6 +495,7 @@ class DatabaseHandler:
         This is intended for cases where a full re-import is done.
         Note that no journal records are kept of this change itself.
         """
+        self._check_write_permitted()
         self._flush_rpsl_object_writing_buffer()
         table = RPSLDatabaseObject.__table__
         stmt = table.delete(table.c.source == source)
@@ -476,6 +515,7 @@ class DatabaseHandler:
         Delete all ROA objects from the database.
         ROAs are always imported in bulk.
         """
+        self._check_write_permitted()
         self._roa_insert_buffer = []
         stmt = ROADatabaseObject.__table__.delete()
         self._connection.execute(stmt)
@@ -486,6 +526,7 @@ class DatabaseHandler:
         Upon the next mirror update, all data for the source will be
         discarded and reloaded from the source.
         """
+        self._check_write_permitted()
         table = RPSLDatabaseStatus.__table__
         synchronised_serials = is_serial_synchronised(self, source, settings_only=True)
         stmt = table.update().where(table.c.source == source).values(
@@ -502,12 +543,14 @@ class DatabaseHandler:
         """
         Record that a mirror was updated to a certain serial.
         """
+        self._check_write_permitted()
         self.status_tracker.record_serial_newest_mirror(source, serial)
 
     def record_serial_seen(self, source: str, serial: int):
         """
         Record that a serial was seen for a source
         """
+        self._check_write_permitted()
         self.status_tracker.record_serial_seen(source, serial)
 
     def record_mirror_error(self, source: str, error: str) -> None:
@@ -515,17 +558,25 @@ class DatabaseHandler:
         Record an error seen in a mirrored database.
         Only the most recent error is stored in the DB status.
         """
+        self._check_write_permitted()
         self.status_tracker.record_mirror_error(source, error)
 
     def record_serial_exported(self, source: str, serial: int) -> None:
         """
         Record an export of a source at a particular serial.
         """
+        self._check_write_permitted()
         self.status_tracker.record_serial_exported(source, serial)
 
     def close(self) -> None:
         self._connection.close()
 
+    def _check_write_permitted(self):
+        if self.readonly:
+            msg = 'Attempted to write to SQL database from readonly database handler'
+            logger.critical(msg)
+            raise Exception(msg)
+
 
 class DatabaseStatusTracker:
     """
